# Copyright 2021 Institute of Plasma Physics of the Czech Academy of Sciences. 
#
# Licensed under the EUPL-1.2 or later.
"""
Contains various tools to ease using the package.

 - DataHandler class prototype
 - projection space plot of a system
"""
from .projection import projection_space
from .data import DataHandler
